/* -*- Mode: vala; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */

/*
  Version 3: json-glib version

  PLAN:
  * add support for GObject Introspection to allow us to serialise non-property members

  json-glib
  * has functions to convert XML structures into Objects and vice versa
  * can convert simple objects automatically
  * richer objects need to implement interface

  json_serializable_real_serialize -> json_serialize_pspec
  * how do these get used with GInterfaces?  are these default methods like with superclasses?
  * TODO: I don't think vala does multiple inheritance, so do we want GXml.Serializable to be an interface or a superclass?

  json_serializable_default_{de,}serialize_property -> json_serializable_real_{de,}serialize

  
  json_serializable_{de,}serialize_property -> iface->{de,}serialize_property
    these all get init'd to -> json_serializable_real_{de,}serialize_property
      these all call -> json_{de,}serialize_pspec

  json_serializable_{find,list,get,set}_propert{y,ies} -> iface->{find,list,get,set}_propert{y,ies}
    these all get init'd to -> json_serializable_real_{find,list,get,set}_propert{y,ies}
	  these all call -> g_object_{class,}_{find,list,get,set}_propert{y,ies}
 */

using GXmlDom;

[CCode (gir_namespace = "GXmlDom", gir_version = "0.2")]
namespace GXmlDom {
	public class Serializer : GLib.Object {
		/* TODO: so it seems we can get property information from GObjectClass
		   but that's about it.  Need to definitely use introspection for anything
		   tastier */
		public GXmlDom.XNode serialize_object (GLib.Object object) {
			/* Create an XML Document to return the object
			   in.  TODO: consider just returning an
			   <Object> node; but then we'd probably want
			   a separate document for it to already be a
			   part of as its owner_document. */
			Document doc = new Document ();
			Element root = doc.create_element ("Object");
			doc.append_child (root);
			root.set_attribute ("otype", object.get_type ().name ());

			/* TODO: make sure we don't use an out param for our returned list
			   size in our interface's list_properties (), using
			   [CCode (array_length_type = "guint")] */
			ParamSpec[] prop_specs = object.get_class ().list_properties ();

			/* Exam the properties of the object and store
			   them with their name, type and value in XML
			   Elements.  Use GValue to convert them to
			   strings. (Too bad deserialising isn't that
			   easy w.r.t. string conversion.) */
			foreach (ParamSpec prop_spec in prop_specs) {
				Element prop = doc.create_element ("Property");
				prop.set_attribute ("ptype", prop_spec.value_type.name ());
				prop.set_attribute ("pname", prop_spec.name);

				Value value = new Value (typeof (string));
				object.get_property (prop_spec.name, ref value);
				prop.content = value.get_string ();
				root.append_child (prop);
			}

			/* Debug output */
			bool debug = false;
			if (debug) {
				stdout.printf ("Object XML\n---\n%s\n", doc.to_string ());

				stdout.printf ("object\n---\n");
				stdout.printf ("get_type (): %s\n", object.get_type ().name ());
				stdout.printf ("get_class ().get_type (): %s\n", object.get_class ().get_type ().name ());
				
				ParamSpec[] properties = object.get_class ().list_properties ();
				stdout.printf ("object has %d properties\n", properties.length);
				foreach (ParamSpec prop in properties) {
					stdout.printf ("---\n");
					stdout.printf ("name: %s\n", prop.name);
					stdout.printf ("value_type: %s\n", prop.value_type.name ());
					stdout.printf ("owner_type: %s\n", prop.owner_type.name ());
					stdout.printf ("get_name (): %s\n", prop.get_name ());
					stdout.printf ("get_blurb (): %s\n", prop.get_blurb ());
					stdout.printf ("get_nick (): %s\n", prop.get_nick ());
				}
			}

			return doc;
		}

		public GLib.Object deserialize_object (XNode node) {
			Element obj_elem = (Element)node;

			NodeList properties = obj_elem.get_elements_by_tag_name ("Property");
			Parameter[] parameters;
			string[] names;

			parameters = new Parameter[properties.length];
			names = new string[properties.length]; // because Parameter.name is unowned

			for (int i = 0; i < properties.length; i++) {
				Element prop_elem = (Element)properties.nth (i);
				Value val = Value (Type.from_name (prop_elem.get_attribute ("ptype")));

				string_to_gvalue (prop_elem.content, ref val);
				names[i] = prop_elem.get_attribute ("pname");
				parameters[i] = { names[i], val };
			}

			Object obj = Object.newv (Type.from_name (obj_elem.get_attribute ("otype")),
						  parameters);

			return obj;
		}

		/* TODO:
		 * - can't seem to pass delegates on struct methods to another function :(
		 * - no easy string_to_gvalue method in GValue :(
		 */
		public static bool string_to_gvalue (string str, ref GLib.Value dest) {
			Type t = dest.type ();
			GLib.Value dest2 = Value (t);
			bool ret = false;
			
			if (t == typeof (int64)) {
				int64 val;
				if (ret = int64.try_parse (str, out val)) {
					dest2.set_int64 (val);
				}
			} else if (t == typeof (int)) {
				int64 val;
				if (ret = int64.try_parse (str, out val)) {
					dest2.set_int ((int)val);
				}
			} else if (t == typeof (long)) {
				int64 val;
				if (ret = int64.try_parse (str, out val)) {
					dest2.set_long ((long)val);
				}
			} else if (t == typeof (uint)) {
				uint64 val;
				if (ret = uint64.try_parse (str, out val)) {
					dest2.set_uint ((uint)val);
				}
			} else if (t == typeof (ulong)) {
				uint64 val;
				if (ret = uint64.try_parse (str, out val)) {
					dest2.set_ulong ((ulong)val);
				}
			} else if (t == typeof (bool)) {
				bool val;
				if (ret = bool.try_parse (str, out val)) {
					dest2.set_boolean (val);
				} 
			} else if (t == typeof (float)) {
				double val;
				if (ret = double.try_parse (str, out val)) {
					dest2.set_float ((float)val);
				}
			} else if (t == typeof (double)) {
				double val;
				if (ret = double.try_parse (str, out val)) {
					dest2.set_double (val);
				}
			} else if (t == typeof (string)) {
				dest2.set_string (str);
				ret = true;
			} else if (t == typeof (char)) {
				int64 val;
				if (ret = int64.try_parse (str, out val)) {
					dest2.set_char ((char)val);
				} 
			} else if (t == typeof (uchar)) {
				int64 val;
				if (ret = int64.try_parse (str, out val)) {
					dest2.set_uchar ((uchar)val);
				}
			} else if (t == Type.BOXED) {
			} else if (t.is_enum ()) {
			} else if (t.is_flags ()) {
			} else if (t.is_object ()) {
			} else {
			}

			if (ret == true) {
				dest = dest2;
				return true;
			} else {
				GLib.warning ("Cannot convert from string to type '%s' yet.", t.to_string ());
				return false;
			}
		}
	}

	public interface SerializableInterface : GLib.Object {
		public abstract void deserialize_property (string property_name, out GLib.Value value, GLib.ParamSpec spec, GXmlDom.XNode property_node);
		public abstract GXmlDom.XNode serialize_property (string property_name, GLib.Value value, GLib.ParamSpec spec);

		/* Correspond to: g_object_class_{find_property,list_properties} */
		public abstract GLib.ParamSpec find_property (string property_name);
		public abstract GLib.ParamSpec[] list_properties (out int num_specs); // TODO: probably unowned

		/* Correspond to: g_object_{set,get}_property */
		public abstract void get_property (GLib.ParamSpec spec, GLib.Value value); // TODO: const?
		public abstract void set_property (GLib.ParamSpec spec, GLib.Value value); // TODO: const?
	}
	
	/**
	 * SECTION:gxml-serializable
	 * @short-description: Serialize and deserialize GObjects
	 *
	 * TODO: elaborate
	 */
	public class Serializable : SerializableInterface, GLib.Object {
		void deserialize_property (string property_name, out GLib.Value value, GLib.ParamSpec spec, GXmlDom.XNode property_node) {
			// TODO: mimic json_deserialize_pspec
			// TODO: consider returning GLib.Value instead of using an out param?

			// convert from XML to a GLib.Value
			return;
		}
		GXmlDom.XNode serialize_property (string property_name, GLib.Value value, GLib.ParamSpec spec) {
			// TODO: mimic json_serialize_pspec

			// convert from GLib.Value to XML
			Type t = value.type ();

			if (t == typeof (int64)) {
				// do this
			} else if (t == typeof (bool)) {
			} else if (t == typeof (double)) {
			} else if (t == typeof (string)) {
			} else if (t == typeof (int)) {
			} else if (t == typeof (float)) {
			} else if (t == Type.BOXED) {
			} else if (t == typeof (uint)) {
			} else if (t == typeof (long)) {
			} else if (t == typeof (ulong)) {
			} else if (t == typeof (char)) {
			} else if (t == typeof (uchar)) {
			} else if (t.is_enum ()) {
			} else if (t.is_flags ()) {
			} else if (t.is_object ()) {
			//} else if (t == typeof (none)) {
			} else {
				
			}

						
			// switch (value.type ().name ()) {
			// case "int64":
			// case "boolean":
			// case "double":
			// case "string":
			// case "int":
			// case "float":
			// case "boxed":
			// case "uint":
			// case "long":
			// case "ulong":
			// case "char":
			// case "uchar":
			// case "enum":
			// case "flags":
			// case "object":
			// case "none":
			// default: /* unsupported */
				
			// }

			return null;
		}

		// g_object_class_{find_property,list_properties}
		GLib.ParamSpec find_property (string property_name) {
			// TODO: call object find property
			return null;
		}
		GLib.ParamSpec[] list_properties (out int num_specs) {
			// TODO: call object list properties
			return null;

		}
		// g_object_{set,get}_property
		void get_property (GLib.ParamSpec spec, GLib.Value value) {
			// TODO: call object get property
		}
		void set_property (GLib.ParamSpec spec, GLib.Value value) {
			// TODO: call object set property
		}
	}
}